/*
Library for the file IO at the KOSMA telescope (and elsewhere ?).
Files are read, their contents searched for 'known' (i.e. declared in
the calling routine) variables. Values are read with the appropriate
data type and stored in the variable's structure member variable.value.
Variables contained in a common file are pointed to by the entries in
the var_set structure, which is used for reading and writing files.

Written by Urs Graf, Feb. 2000

$Id: KOSMA_file_io.c,v 1.4 2001/07/13 11:18:46 driveprogram Exp $

Todos:
*/

#include "KOSMA_file_io.h"
#include "KOSMA_file_io_internal.h"

static VAR_SET_LIST read_var_set_list = { NULL, NULL };
static VAR_SET_LIST write_var_set_list = { NULL, NULL };
static char col_sep[] = { ' '  ,  ','  ,  ';'  ,  '\t'  ,  '\n' };
static char *config_file_name = NULL;
static char *config_dir = NULL;
static char *read_dir = NULL;
static char *write_dir = NULL;
static int init = TRUE;
static char dimension_string[ INPUTSTRINGLENGTH ];



void Initialize()
{
  /* set directories */
  if ( !config_file_name )
    DefineConfigFile( "KOSMA_file_io.cfg" );
  if ( !config_dir )
    SetConfigDir( getenv( "CONFIG_DIR" ) );
  if ( !read_dir )
    SetReadDir( getenv( "READ_DIR" ) );
  if ( !write_dir )
    SetWriteDir( getenv( "WRITE_DIR" ) );

  init = FALSE;

  return;
}



/* Configure an int variable ( or an array - controlled by dim-string ) */
int ConfigureInt( INT_VAR *var, char *name, char *filename, char *dim )
{
  int (*function)() = ConfigureInt;
  static INT_VAR *variable;
  static char basename[ INPUTSTRINGLENGTH ];
  char errorstring[ERRORSTRINGLENGTH];

  if ( var )
  {
    /* function was called from main program, not from recursion */
    variable = var;
    strcpy( basename, name );
  }

  if ( *dim )
    /* append name extensions ( i.e. [2][1][4] ) to array elements 
       through a recursive function call */
    GenerateArrayExtension( name, filename, function, dim );
  else
  {
    /* recursion has reduced dim-string to "", we are now dealing with 
       a scalar and can start doing things */
    variable->read_fct = ReadIntVar;
    variable->write_fct = WriteIntVar;
    variable->value = 0;
    ConfigureGeneric( (GENERIC_VAR *)variable, basename, filename );
    /* reset variable name from basename to full name */
    variable->name = (char *)malloc( strlen( name ) + 1 );
    strcpy( variable->name, name );

    sprintf( errorstring, "Variable %s configured", variable->name );
    SetError( INFO, errorstring );
    ErrorMessage();

    variable++;
  }

  return( 0 );
}



/* Configure a double variable ( or an array - controlled by dim-string ) */
int ConfigureDouble( DOUBLE_VAR *var, char *name, char *filename, char *dim )
{
  int (*function)() = ConfigureDouble;
  static DOUBLE_VAR *variable;
  static char basename[ INPUTSTRINGLENGTH ];
  char errorstring[ERRORSTRINGLENGTH];

  if ( var )
  {
    /* function was called from main program, not from recursion */
    variable = var;
    strcpy( basename, name );
  }

  if ( *dim )
    /* append name extensions ( i.e. [2][1][4] ) to array elements 
       through a recursive function call */
    GenerateArrayExtension( name, filename, function, dim );
  else
  {
    /* recursion has reduced dim-string to "", we are now dealing with 
       a scalar and can start doing things */
    variable->read_fct = ReadDoubleVar;
    variable->write_fct = WriteDoubleVar;
    variable->value = 0;
    ConfigureGeneric( (GENERIC_VAR *)variable, basename, filename );
    /* reset variable name from basename to full name */
    variable->name = (char *)malloc( strlen( name ) + 1 );
    strcpy( variable->name, name );

    sprintf( errorstring, "Variable %s configured", variable->name );
    SetError( INFO, errorstring );
    ErrorMessage();

    variable++;
  }

  return( 0 );
}



/* Configure a string variable ( or an array - controlled by dim-string ) */
int ConfigureString( STRING_VAR *var, char *name, char *filename, char *dim )
{
  int (*function)() = ConfigureString;
  static STRING_VAR *variable;
  static char basename[ INPUTSTRINGLENGTH ];
  char errorstring[ERRORSTRINGLENGTH];

  if ( var )
  {
    /* function was called from main program, not from recursion */
    variable = var;
    strcpy( basename, name );
  }

  if ( *dim )
    /* append name extensions ( i.e. [2][1][4] ) to array elements 
       through a recursive function call */
    GenerateArrayExtension( name, filename, function, dim );
  else
  {
    /* recursion has reduced dim-string to "", we are now dealing with 
       a scalar and can start doing things */
    variable->read_fct = ReadStringVar;
    variable->write_fct = WriteStringVar;
    variable->value = NULL;
    ConfigureGeneric( (GENERIC_VAR *)variable, basename, filename );
    /* reset variable name from basename to full name */
    variable->name = (char *)malloc( strlen( name ) + 1 );
    strcpy( variable->name, name );

    sprintf( errorstring, "Variable %s configured", variable->name );
    SetError( INFO, errorstring );
    ErrorMessage();

    variable++;
  }

  return( 0 );
}



/* append name extensions ( i.e. [2][1][4] ) to array elements 
   through a recursive function call. 
   Here it is important to iterate through the array in the same sequence
   used by C for storing multidimensinal arrays, because the calling
   routine is just incrementing a pointer for each element */
int GenerateArrayExtension( char *name, char *filename,
                            int (*function)(), char *dim )
{
  int i, n;
  char *new_dim, new_name[INPUTSTRINGLENGTH];

  /* read max index of first dimension from dim-string */
  sscanf( dim, "%d", &n );

  /* scan dim-string for next value and set new_dim pointer to it */
  new_dim = dim + 1;
  /* first looking for a whitespace character */
  while( !isspace( new_dim[0] ) && *new_dim != 0 )
    new_dim = new_dim + 1;
  /* then looking for a number */
  while( !isdigit( new_dim[0] ) && *new_dim != 0 )
    new_dim = new_dim + 1;

  /* loop through current index */
  for ( i = 0; i < n; i++ )
  {
    /* generate extension string */
    sprintf( new_name, "%s[%d]", name, i );
    /* recursive call to calling function */
    /* NULL pointer tells the calling routine that we are inside the 
       recursive loop */
    function( NULL, new_name, filename, new_dim );
  }

  return( 0 );
}



int ConfigureGeneric( GENERIC_VAR *variable, char *name, char *filename )
{
  FILE *configfile;
  char inputstring[INPUTSTRINGLENGTH], errorstring[ERRORSTRINGLENGTH];
  int found = FALSE;
  char *fullfilename;

  if ( init )
    Initialize();

  /* Standard initialization of structure */
  variable->name = name;
  variable->read_var_set = NULL;
  variable->write_format = NULL;
  variable->write_var_set_list.first = NULL;
  variable->write_var_set_list.last = NULL;
  variable->comment = NULL;

  if ( !*filename )
    filename = config_file_name;
  fullfilename = AddPath( config_dir, filename );
  if (( configfile = fopen( fullfilename, "r" ) ) == NULL) /* open file */
  {  
    /* file not found */
    sprintf( errorstring, "Could not open configuration file %s", 
                          fullfilename );
    SetError( FATAL, errorstring );
    ErrorMessage();

    exit( FATAL );
  }


  /* read config file */
  while ( fgets( inputstring, INPUTSTRINGLENGTH - 1, configfile ) != NULL)
  {
    /* try to find config file entry for variable */
    if ( ParseConfigLine1( variable, inputstring ) )
    {
      found = TRUE;
      /* got it, now read second config line */
      if ( fgets( inputstring, INPUTSTRINGLENGTH - 1, configfile ) == NULL )
      {
        /* second config line does not exist */
        sprintf( errorstring, 
          "Second config line for variable %s does not exist", variable->name );
        SetError( FATAL, errorstring );
        ErrorMessage();

        exit( FATAL );
      }
      if( ParseConfigLine2( variable, inputstring ) )
        break;
      else
      {
        sprintf( errorstring, 
          "Could not configure variable %s", variable->name );
        SetError( FATAL, errorstring );
        ErrorMessage();

        exit( FATAL ); /* */
      }
    }
  }

  fclose( configfile );

  if ( !found )
  {
    /* could not find config data */
    sprintf( errorstring, "Could not find %s in config file %s", 
                          variable->name, filename );
    SetError( FATAL, errorstring );
    ErrorMessage();
    exit ( FATAL );
  }
  else
    return( 0 );

}



int ParseConfigLine1( GENERIC_VAR *variable, char *string )
{
  char *name_ptr, *format_ptr, *value_ptr;

  value_ptr = string;

  /* search for variable name in inputstring */
  if ( !SplitInputString( string, variable->name, &name_ptr, &format_ptr ) )
    /* name string not found in inputstring */
    return( FALSE ); 

  /* name string found in inputstring */

  /* now read default value */
  variable->read_fct( variable, value_ptr );
  /* read format string */
  variable->write_format = 
    ReplaceString( variable->write_format, TrimString( format_ptr ) );

  return( TRUE );
}



/* search for a white space delimited version of marker in string
   return pointers to begin and end(+1) of marker. Replace white space 
   in front of marker by \0 to terminate first part of string */
int SplitInputString( char *string, char *marker, char **sta, char **sto )
{
  char *start, *stop, errorstring[ERRORSTRINGLENGTH];
  int length, found = FALSE;
  
  length = strlen( marker );
  start = string;
  while ( *start != LINEFEED )
  {
    if ( *start == COMMENT_TAG ) /* ignore everything that is in a comment */
      return( FALSE );

    if ( *start == *marker )
    {
      /* search for variable's name string in inputstring */
      if ( !strncmp( start, marker, length ) )
      {
        /* name string found */
        stop = start + length;
        /* check whether it is white space terminated */
        if ( isspace( *stop ) && isspace( start[-1] ) )
        {
          found = TRUE;
          start[-1] = 0; /* terminate string section in front of marker */
          break;
        }
        else
        {
          sprintf( errorstring, 
                   "String %s found, but not white space delimited", 
                   marker );
          SetError( DEBUG, errorstring );
          ErrorMessage();
        }
      }
    }
    start++;
  }

  *sta = start;
  *sto = stop;

  return ( found );
}




int ParseConfigLine2( GENERIC_VAR *variable, char *string )
{
  char *readfile_ptr, *writefile_ptr;
  VAR_SET_LIST_ENTRY *var_set_list_entry;
  VAR_SET *write_var_set;
  char errorstring[ERRORSTRINGLENGTH];
  int found = FALSE;

  /* get comment string from second line of configuration record */
  variable->comment = ReplaceString( variable->comment, 
                      TrimString( GetCommentString( string ) ) );

  /* get read-file name */
  readfile_ptr = strtok( string, col_sep );

  if ( !readfile_ptr ) /* nothing left in inputstring */
  {
    variable->read_fct = IllegalRead;
    variable->write_fct = IllegalWrite;
    sprintf( errorstring, "Neither input nor output file configured for %s",
             variable->name );
    SetError( WARNING, errorstring );
    ErrorMessage();

    return( 1 );
  }

  /* configure read_var_set */
  if ( !strcmp( readfile_ptr, "NULL" ) )
    variable->read_fct = IllegalRead;  /* write only variable */
  else
  {
    /* check if var_set exists already */
    var_set_list_entry = read_var_set_list.first;
    while ( var_set_list_entry )
    {
      if ( !strcmp( readfile_ptr, var_set_list_entry->var_set->filename ) )
      {
        /* read_var_set found in list */
        variable->read_var_set = var_set_list_entry->var_set;
        AppendToVarSet( &(variable->name), variable->read_var_set );
        break;
      }
      var_set_list_entry = var_set_list_entry->next;
    }
  
    if ( !variable->read_var_set )
    {
      /* read_var_set does not exist yet. Create one */
      variable->read_var_set = CreateReadVarSet( readfile_ptr );
      AppendToVarSet( &(variable->name), variable->read_var_set );
    }
  } 


  /* configure write_var_sets */
  /* get write-file name from input string */
  while ( ( writefile_ptr = strtok( NULL, col_sep ) ) )
  {
    if ( !strcmp( writefile_ptr, "NULL" ) )
    {
      variable->write_fct = IllegalWrite;
      /* read-only variable, no need to go on */
      return( 1 );
    }

    /* check if var_set exists already */
    var_set_list_entry = write_var_set_list.first;
    while ( var_set_list_entry )
    {
      if ( !strcmp( writefile_ptr, var_set_list_entry->var_set->filename ) )
      {
        /* write_var_set found in list */
        found = TRUE;
        AppendToVarSetList( var_set_list_entry->var_set,
                            &(variable->write_var_set_list) );
        AppendToVarSet( &(variable->name), var_set_list_entry->var_set );
        break;
      }
      var_set_list_entry = var_set_list_entry->next;
    }

    if ( !found )
    {
      /* write_var_set does not exist yet. Create one */
      write_var_set = CreateWriteVarSet( writefile_ptr );
      AppendToVarSetList( write_var_set, &(variable->write_var_set_list) );
      AppendToVarSet( &(variable->name), write_var_set );
    }

    found = FALSE; /* prepare for next file name */
  }

  return( TRUE );
}



VAR_SET *CreateReadVarSet( char *filename )
{
  VAR_SET *var_set;

  var_set = CreateVarSet( filename );
  AppendToVarSetList( var_set, &read_var_set_list );

  return( var_set );
}



VAR_SET *CreateWriteVarSet( char *filename )
{
  VAR_SET *var_set;

  var_set = CreateVarSet( filename );
  AppendToVarSetList( var_set, &write_var_set_list );

  return( var_set );
}



VAR_SET *CreateVarSet( char *filename )
{
  VAR_SET *var_set;
  STRING_VAR *variable;

  var_set = (VAR_SET *)malloc( sizeof( VAR_SET ) );
  var_set->filename = (char *)malloc( strlen( filename ) + 1 );
  strcpy( var_set->filename, filename );

  /* initialize var_set with time marker as first entry */
  var_set->first = (VAR_SET_ENTRY *)malloc( sizeof( VAR_SET_ENTRY ) );
  var_set->last = var_set->first;

  /* initialize time marker entry */
  var_set->first->next = NULL;
  variable = (STRING_VAR *)malloc( sizeof( STRING_VAR ) );
  var_set->first->variable = &variable->name;
  variable->name = TIMESTAMPNAME;
  variable->comment = (char *)malloc( 1 );
  *variable->comment = 0;
  variable->read_var_set = var_set;
  variable->read_fct = ReadStringVar;
  variable->write_var_set_list.first = (VAR_SET_LIST_ENTRY *)var_set;
  variable->write_var_set_list.last = NULL;
  variable->write_fct = WriteStringVar;
  variable->write_format = "%s";
  variable->value = (char *)calloc( INPUTSTRINGLENGTH, 1 );
  memset( variable->value, '0', INPUTSTRINGLENGTH );

  return( var_set );
}



char *CreateTimeMark( char *timestring )
{
  char date[INPUTSTRINGLENGTH];
  struct timeval tv;
  struct timezone tz;

  gettimeofday( &tv, &tz );
  strftime( date, TIMESTRINGLENGTH + 1, "%d-%b-%Y  %H:%M:%S",
                                        localtime( &tv.tv_sec ) );

  sprintf( date + TIMESTRINGLENGTH, "%12d.%06d", 
                                    (int)tv.tv_sec, (int)tv.tv_usec );

  timestring = ReplaceString( timestring, date );

  return( timestring );
}



void AppendToVarSetList( VAR_SET *var_set, VAR_SET_LIST *var_set_list )
{
  VAR_SET_LIST_ENTRY  *entry_ptr, *newentry_ptr;

  newentry_ptr = (VAR_SET_LIST_ENTRY *)malloc( sizeof( VAR_SET_LIST_ENTRY ) );
  newentry_ptr->var_set = var_set;
  newentry_ptr->next = NULL;

  /* find last entry of var_set_list */
  if ( var_set_list->first )
  {
    entry_ptr = var_set_list->first;
    while ( entry_ptr->next )
    {
      entry_ptr = entry_ptr->next;
    }
    entry_ptr->next = newentry_ptr;
  }
  else
  {
    var_set_list->first = newentry_ptr;
  }

  var_set_list->last = newentry_ptr;


  return;
}



void AppendToVarSet( VAR_NAME *variable, VAR_SET *var_set )
{
  VAR_SET_ENTRY  *entry_ptr, *newentry_ptr;

  newentry_ptr = (VAR_SET_ENTRY *)malloc( sizeof( VAR_SET_ENTRY ) );
  newentry_ptr->variable = variable;
  newentry_ptr->next = NULL;

  /* find last entry of var_set */
  if ( var_set->first )
  {
    entry_ptr = var_set->first;
    while ( entry_ptr->next )
    {
      entry_ptr = entry_ptr->next;
    }
    entry_ptr->next = newentry_ptr;
  }
  else
  {
    var_set->first = newentry_ptr;
  }

  var_set->last = newentry_ptr;

  return;
}



int ParseInputLine( GENERIC_VAR *variable, char *string )
{
  char *name_ptr = NULL, *value_ptr, *comment_ptr;

  value_ptr = string;
  
  /* search for variable name in inputstring */
  if ( !SplitInputString( string, variable->name, &name_ptr, &comment_ptr ) )
    /* name string not found in inputstring */
    return( FALSE ); 

  /* name string found in inputstring */

  /* now read value */
  variable->read_fct( variable, value_ptr );
  /* read comment string */
  variable->comment = ReplaceString( variable->comment,
                      TrimString( GetCommentString( comment_ptr ) ) );

  return( TRUE );
}



char *GetCommentString( char *string )
{
  char *p;

  p = strchr( string, COMMENT_TAG );
  if ( p != NULL )
  /* found start of comment string */
  {
    *p = 0; /* replace comment-tag by null character to
               terminate inputstring before comment starts */
    p++;
  }
  else
  {
    p = strchr( string, LINEFEED );
    *p = 0; /* replace line feed by null character to terminate inputstring */
  }

  return( p );
}



/* read all variables registered in global read_var_set_list */
double ReadAllVariables()
{
  VAR_SET_LIST_ENTRY var_set_list_entry;
  double newdata = -1., newdatasum = 0., errorsum = 0.;

  var_set_list_entry.next = read_var_set_list.first;
  while ( var_set_list_entry.next )
  {
    var_set_list_entry = *var_set_list_entry.next;
    newdata = ReadVarSet( var_set_list_entry.var_set );
    if ( newdata >= 0 )
      newdatasum = newdatasum + newdata;
    else 
      errorsum = errorsum + newdata;
  }

  if ( errorsum != 0. )
    return( errorsum );
  else
    return( newdatasum );
}



/* Read a variable's var_set from file (i.e. all variables in the same set) */
double ReadVariablesVarSet( VAR_NAME *var_name )
{
  GENERIC_VAR *variable;

  variable = (GENERIC_VAR *)var_name;

  if ( variable->read_var_set )
    return( ReadVarSet( variable->read_var_set ) );
  else
  {
    variable->read_fct( &variable->name );
    return( -1. );
  }
}



/* Read var_set from file. */
double ReadVarSet( VAR_SET *var_set )
{
  char errorstring[ERRORSTRINGLENGTH];
  FILE *readfile;
  GENERIC_VAR *variable;
  VAR_SET_ENTRY *var_set_entry;
  double newdata;
  int readerror, globalreaderror = 0;

  typedef struct inputstring
  {
    char string[INPUTSTRINGLENGTH];
    struct inputstring *next;
  } INPUTSTRING;
  INPUTSTRING *input_entry, *input, first, *obsolete_entry, *previous;

  /* open file and check timestamp */
  newdata = OpenReadFile( var_set, &readfile );
  if ( newdata <= 0 )
    return( newdata ); /* no new data available */

  /* read file in memory */
  input = ( INPUTSTRING *)malloc( sizeof( INPUTSTRING ) );
  previous = input;
  input->next = NULL;
  first.next = input; /* mark beginning of list */
  while ( fgets(input->string, INPUTSTRINGLENGTH - 1, readfile) != NULL)
  {
    previous = input; /* save pointer to current end of list */
    input->next = ( INPUTSTRING *)malloc( sizeof( INPUTSTRING ) );
    input = input->next;
    input->next = NULL;
  }
  fclose( readfile );
  /* clean up */
  free( previous->next );
  previous->next = NULL;
  /* done reading file */

  /* scan through variable list */
  var_set_entry = var_set->first;
  while ( var_set_entry->next )
  {
    var_set_entry = var_set_entry->next;
    variable = (GENERIC_VAR *)var_set_entry->variable;

    /* prepare error flag in case we do not find the variable */
    readerror = TRUE;

    /* scan through input list */
    input_entry = &first;
    while ( input_entry->next )
    {
      if ( ParseInputLine( variable, input_entry->next->string ) )
      {
        /* variable found and read from input list */
        obsolete_entry = input_entry->next;
        input_entry->next = input_entry->next->next;
        free( obsolete_entry );
        readerror = FALSE; /* reset error flag */
        break;
      }
      input_entry = input_entry->next;
    }
    if ( readerror )
    {
      sprintf( errorstring, "%s not found in input file %s", 
               variable->name, variable->read_var_set->filename );
      SetError( ERROR, errorstring );
      globalreaderror++;
    }
    else
    {
      sprintf( errorstring, "%s read from input file %s", 
               variable->name, variable->read_var_set->filename );
      SetError( INFO, errorstring );
    }
    ErrorMessage();
  }
  /* done scanning through variable list */

  /* clean up */
  if( first.next )
  {
    input_entry = first.next;
    while ( input_entry->next )
    {
      obsolete_entry = input_entry;
      input_entry = input_entry->next;
      free( obsolete_entry );
    }
    free( input_entry );
  }
  /* done cleaning  up */

  if ( globalreaderror )
  {
    sprintf( errorstring, "Input file %s read with errors", 
                          var_set->filename );
    SetError( ERROR, errorstring );
    newdata = -globalreaderror; /* to return the error count */
  }
  else
  {
    sprintf( errorstring, "Input file %s successfully read", 
                          var_set->filename );
    SetError( INFO, errorstring );
  }
  ErrorMessage();

  return( newdata );
}



/* Read single variable. */
double ReadVar( VAR_NAME *variable_name)
{
  char inputstring[INPUTSTRINGLENGTH];
  char errorstring[ERRORSTRINGLENGTH];
  FILE *readfile;
  GENERIC_VAR *variable;
  double newdata;
  int found = FALSE;

  variable = (GENERIC_VAR *)variable_name;
  
  if ( !variable->read_var_set )
  {
    /* attempt explicit reading of write-only variable */
    variable->read_fct( variable );
    return( -1 );
  }

  /* open file and check timestamp */
  newdata = OpenReadFile( variable->read_var_set, &readfile );
  if ( newdata <= 0 )
    return( newdata ); /* no new data available */

  /* new data available */

  /* prepare error message in case we do not find the variable */
  sprintf( errorstring, "%s not found in input file %s", 
           variable->name, variable->read_var_set->filename );
  SetError( ERROR, errorstring );
 
  while ( fgets(inputstring, INPUTSTRINGLENGTH - 1, readfile) != NULL)
  {
    if ( ParseInputLine( variable, inputstring ) )
    {
      /* Found!!!! */
      found = TRUE;
      sprintf( errorstring, "%s read from input file %s", 
               variable->name, variable->read_var_set->filename );
      SetError( INFO, errorstring );
      break;
    }
  }
  fclose( readfile );
  ErrorMessage();

  if ( found )
    return( newdata );
  else
    return( -1. );
}



/* open input file and check time stamp */
double OpenReadFile( VAR_SET *var_set, FILE **readfile ) 
{
  char inputstring[INPUTSTRINGLENGTH];
  char errorstring[ERRORSTRINGLENGTH];
  STRING_VAR *timestamp;
  char *fullfilename;
  double lasttime, thistime = 0., difftime = 10000;

  /* open input file */
  fullfilename = AddPath( read_dir, var_set->filename );
  if (( *readfile = fopen( fullfilename, "r") ) == NULL)
  {  
    /* file not found */
    sprintf( errorstring, "Could not open input file %s", fullfilename );
    SetError( ERROR, errorstring );
    ErrorMessage();

    return( -1. ); /* nothing more to do here */
  }

  /* read first line of file */
  if ( fgets(inputstring, INPUTSTRINGLENGTH - 1, *readfile) == NULL)
  {
    fclose( *readfile );
    sprintf( errorstring, "Input file %s is empty", fullfilename );
    SetError( DEBUG, errorstring );
    ErrorMessage();

    return( -1. );  /* file is empty */
  }

  /* check timestamp */
  timestamp = (STRING_VAR *)var_set->first->variable;
  /* save previous time stamp */
  sscanf( timestamp->value + TIMESTRINGLENGTH, "%lf", &lasttime );

  /* prepare error message in case we do not find the variable */
  sprintf( errorstring, "time stamp missing in first line of input file %s", 
           var_set->filename );
  SetError( DEBUG, errorstring );

  if ( ParseInputLine( (GENERIC_VAR *)timestamp, inputstring ) )
  {
    /* variable found and read from input string */
    sprintf( errorstring, "time stamp found in input file %s", 
             var_set->filename );
    SetError( DEBUG, errorstring );

    /* read timestamp */
    sscanf( timestamp->value + TIMESTRINGLENGTH, "%lf", &thistime );
    difftime = thistime - lasttime;
  }
  else
    rewind( *readfile ); /* to enable the calling routine to reread the 
                            first line of the data file */

  ErrorMessage();
  /* done checking timestamp */

  if ( difftime == 0 )
  {
    sprintf( errorstring, "no new data in input file %s; file not read", 
             var_set->filename );
    SetError( INFO, errorstring );
    ErrorMessage();

    fclose( *readfile );
  }

  return( difftime );
}



/* read value of integer variable from input string */
int ReadIntVar( GENERIC_VAR *var, char *string )
{
  INT_VAR *variable;

  variable = (INT_VAR *)var;
  sscanf( string, "%d", &(variable->value) );

  return( 0 );
}



/* read value of double variable from input string */
int ReadDoubleVar( GENERIC_VAR *var, char *string )
{
  DOUBLE_VAR *variable;

  variable = (DOUBLE_VAR *)var;
  sscanf( string, "%lf", &(variable->value) );

  return( 0 );
}



/* read value of string variable from input string */
int ReadStringVar( GENERIC_VAR *var, char *string )
{
  STRING_VAR *variable;

  variable = (STRING_VAR *)var;
  variable->value = ReplaceString( variable->value, TrimString( string ) );

  return( 0 );
}



/* write all files */
int WriteAllVariables()
{
  return( WriteVarSetList( write_var_set_list ) );
}



/* write all files contained in a write_var_set_list */
int WriteVarSetList( VAR_SET_LIST list )
{
  VAR_SET_LIST_ENTRY *entry_ptr;
  int errorcount = 0, ret_val = 0;

  entry_ptr = list.first;
  while ( entry_ptr->next )
  {
    ret_val = WriteVarSet( *(entry_ptr->var_set) );
    if ( ret_val )
      errorcount = errorcount + ret_val;
    entry_ptr = (VAR_SET_LIST_ENTRY *)entry_ptr->next;
  }
  ret_val = WriteVarSet( *(entry_ptr->var_set) );

  if ( ret_val )
    return( errorcount + ret_val );
  else
    return( ret_val );
}



/* write all variables within a var_set */
int WriteVarSet( VAR_SET var_set )
{
  VAR_SET_ENTRY *entry_ptr;
  GENERIC_VAR *variable;
  STRING_VAR *timestamp;
  FILE *writefile;
  char *fullfilename;
  char errorstring[ERRORSTRINGLENGTH];

  fullfilename = AddPath( write_dir, var_set.filename );
  if (( writefile = fopen( fullfilename, "w" ) ) != NULL) /* open file */
  {
    /* writefile = stdout; just for debugging */
    entry_ptr = var_set.first;
    /* write the current time stamp to the first variable of the var_set */
    timestamp = (STRING_VAR *)entry_ptr->variable;
    timestamp->value = CreateTimeMark( timestamp->value );
    /* scan through variable list */
    while ( entry_ptr->next )
    {
      variable = (GENERIC_VAR *)entry_ptr->variable; /* need to cast 
        the pointer in order to access structure member write_fct */
      variable->write_fct( entry_ptr->variable, writefile );
      entry_ptr = (VAR_SET_ENTRY *)entry_ptr->next;
    }
    variable = (GENERIC_VAR *)entry_ptr->variable;
    variable->write_fct( entry_ptr->variable, writefile );
    
    fclose( writefile );

    sprintf( errorstring, "Output file %s successfully written", 
                           var_set.filename );
    SetError( INFO, errorstring );
    ErrorMessage();

    return( 0 );
  }
  else
  {
    /* file not found */
    sprintf( errorstring, "Could not open output file %s", fullfilename );
    SetError( ERROR, errorstring );
    ErrorMessage();

    return( -1 );
  }
}



/* write a variable to all files contained in variable's write_var_set-list */
int WriteVariablesVarSet( VAR_NAME *variable_name )
{
  GENERIC_VAR *variable;

  variable = (GENERIC_VAR *)variable_name;
  if ( variable->write_fct == IllegalWrite )
    variable->write_fct( variable ); /* that just prints an error message */
  else
    return( WriteVarSetList( variable->write_var_set_list ) );

  return( -1 );
}



/* write a variable to all files contained in variable's write_var_set-list */
int WriteVar( VAR_NAME *variable_name )
{
  return( WriteVariablesVarSet( variable_name ) );
}



/* write variable of type integer pointed to by var_name into writefile */
int WriteIntVar( VAR_NAME *var_name, FILE *writefile )
{
  INT_VAR *variable;

  variable = (INT_VAR *)var_name; /* need to cast the pointer in
    order to access structure members */

  if ( !variable->write_var_set_list.first )
    return( 0 ); /* read only variable */

  fprintf( writefile, variable->write_format, variable->value );
  fprintf( writefile, "   %s   ", variable->name );
  fprintf( writefile, "%c %s\n", COMMENT_TAG, variable->comment );

  return( 0 );
}



/* write variable of type double pointed to by var_name into writefile */
int WriteDoubleVar( VAR_NAME *var_name, FILE *writefile )
{
  DOUBLE_VAR *variable;

  variable = (DOUBLE_VAR *)var_name; /* need to cast the pointer in
    order to access structure members */

  if ( !variable->write_var_set_list.first )
    return( 0 ); /* read only variable */

  fprintf( writefile, variable->write_format, variable->value );
  fprintf( writefile, "   %s   ", variable->name );
  fprintf( writefile, "%c %s\n", COMMENT_TAG, variable->comment );

  return( 0 );
}



/* write string variable pointed to by var_name into writefile */
int WriteStringVar( VAR_NAME *var_name, FILE *writefile )
{
  STRING_VAR *variable;

  variable = (STRING_VAR *)var_name; /* need to cast the pointer in
    order to access structure members */

  if ( !variable->write_var_set_list.first )
    return( 0 ); /* read only variable */

  fprintf( writefile, variable->write_format, variable->value );
  fprintf( writefile, "   %s   ", variable->name );
  fprintf( writefile, "%c %s\n", COMMENT_TAG, variable->comment );

  return( 0 );
}



/* trim leading and trailing blanks from string */
char *TrimString( char *string )
{
  char *end_ptr;

  /* trim leading blanks from string */
  while ( isspace( *string ) )
    string++;
  /* trim trailing blanks from string */
  end_ptr = string + strlen( string ) - 1;
  while ( isspace( *end_ptr ) )
  {
    *end_ptr = 0;
    end_ptr--;
  }

  return( string );
}



/* free memory pointed to by p, allocate memory for a copy of the string 
   and return the pointer to the copy */
char *ReplaceString( char *p, char *string )
{
  /* free previously allocated memory */
  if ( p )
    free( p );
  /* allocate space for a copy */
  p = (char *)malloc( strlen( string ) + 1 );
  /* "+1" because strcpy includes the ending 0 */
  strcpy( p, string );

  return( p );
}



/* refuse to read write-only variable */
int IllegalRead( VAR_NAME *var_name )
{
  GENERIC_VAR *variable;
  char errorstring[ERRORSTRINGLENGTH];

  variable = (GENERIC_VAR *)var_name;
  sprintf( errorstring, "Trying to read write-only variable %s", 
                        variable->name );
  SetError( ERROR, errorstring );
  ErrorMessage();

  return( -1 );
}



/* refuse to write read-only variable */
int IllegalWrite( VAR_NAME *var_name )
{
  GENERIC_VAR *variable;
  char errorstring[ERRORSTRINGLENGTH];

  variable = (GENERIC_VAR *)var_name;
  sprintf( errorstring, "Trying to write read-only variable %s", 
                        variable->name );
  SetError( WARNING, errorstring );
  ErrorMessage();

  return( -1 );
}



/* set default name for config file */
void DefineConfigFile( char *file_name )
{
  char errorstring[ERRORSTRINGLENGTH];

  config_file_name = ReplaceString( config_file_name, file_name );

  sprintf( errorstring, "config file name set to %s", config_file_name );
  SetError( INFO, errorstring );
  ErrorMessage();

  return;
}



/* set standard name for config file */
void StandardConfigFile( char *file_name )
{
  char errorstring[ERRORSTRINGLENGTH];
  char string[INPUTSTRINGLENGTH];


  /* add "fg" to file name to turn file name extension "c" into "cfg" */
  sprintf( string, "%s%s", file_name, "fg" );
  config_file_name = ReplaceString( config_file_name, string );

  sprintf( errorstring, "config file name set to %s", config_file_name );
  SetError( INFO, errorstring );
  ErrorMessage();

  return;
}



/* set directory path for config file */
void SetConfigDir( char *dir_name )
{
  char errorstring[ERRORSTRINGLENGTH];

  config_dir = dir_name;
  if ( !config_dir )
    config_dir = "";
  sprintf( errorstring, "config-path set to %s", dir_name );
  SetError( INFO, errorstring );
  ErrorMessage();

  return;
}



/* set directory path for read file */
void SetReadDir( char *dir_name )
{
  char errorstring[ERRORSTRINGLENGTH];

  read_dir = dir_name;
  if ( !read_dir )
    read_dir = "";
  sprintf( errorstring, "read-path set to %s", dir_name );
  SetError( INFO, errorstring );
  ErrorMessage();

  return;
}



/* set directory path for write file */
void SetWriteDir( char *dir_name )
{
  char errorstring[ERRORSTRINGLENGTH];

  write_dir = dir_name;
  if ( !write_dir )
    write_dir = "";
  sprintf( errorstring, "write-path set to %s", dir_name );
  SetError( INFO, errorstring );
  ErrorMessage();

  return;
}



/* set directory path for write file */
char *AddPath( char *dir, char *file )
{
  static char fullfilename[INPUTSTRINGLENGTH];
  
  memset( fullfilename, 0, INPUTSTRINGLENGTH );
  if( *dir )
  {
    strcat( fullfilename, dir );
    strcat( fullfilename, "/" );
  }
  strcat( fullfilename, file );

  return( fullfilename );
}



/* concatenate non-zero integers to a blank separated string */
char *Dimension( int n, ... )
{
  va_list arg_list;
  char string[ INPUTSTRINGLENGTH ];

  dimension_string[0] = 0; /* needs to be global, since we return it */

  va_start( arg_list, n); /* initialize variable argument list */

  /* as long as we have non-zero arguments ... */
  while ( n )
  {
    sprintf( string, " %d", n );
    /* check whether there is enough space for concatenation */
    if ( strlen( dimension_string ) + strlen( string ) + 2 > INPUTSTRINGLENGTH )
    {
      SetError( ERROR, "Dimension string too long. Prematurely terminated" );
      ErrorMessage();
      break;
    }
    strcat( dimension_string, string );
    /* get next argument */
    n =  va_arg( arg_list, int );
  }

  va_end( arg_list ); /* close variable argument list */

  return( dimension_string );
}
