/*
Simple error message handler. 

Consists of the following routines:
- void DebugLevel( int dl )  select your debug level
- void ErrorLogfile( const char *filename )  select a logfile name
                                             "" for no error logfile
- void SetError( int el, const char *string )  submit an error level/message
- void ErrorMessage()  display and log error message

Written by Urs Graf, Jan. 2001

$Id: error.c,v 1.2 2001/07/13 11:18:46 driveprogram Exp $

*/

#include "KOSMA_error.h"

#define FILENAMELENGTH 257

static char error_logfile[ FILENAMELENGTH ];
static int debuglevel = DEBUGLEVEL;
static int logginglevel = DEBUGLEVEL;
static int errorlevel = 0;
static char error_message[ ERRORSTRINGLENGTH ];
static char filename[ ERRORSTRINGLENGTH ];
static int linenumber;
static const char *errorcode[10] = { "", "DEBUG: ", "INFO: ", "WARNING: ", 
                                     "ERROR: ", "FATAL: ", "", "", "", "" };
static int long_format = FALSE;


/* set debug level */
void DebugLevel( int dl, int ll )
{
  debuglevel = dl;
  if ( debuglevel < 0 )
    debuglevel = 0;

  logginglevel = ll;
  if ( logginglevel < 0 )
    logginglevel = 0;

  return;
}


/* set format of error messages */
void SetErrorFormat( char *ef )
{
  if ( toupper( *ef ) == 'L' )
    long_format = TRUE;
  else
    long_format = FALSE;

  return;
}


/* set logfile name */
void ErrorLogfile( const char *filename )
{
  time_t now;
  char date[23];
  char string[ ERRORSTRINGLENGTH ];
  char new_logfile[ FILENAMELENGTH ];
  FILE *logfile;

  time( &now );
  strftime( date, 22, "%d-%b-%Y  %H:%M:%S", localtime( &now ) );

  if ( error_logfile[0] )
  {
    /* stop logging in old logfile */
    if (( logfile = fopen( error_logfile, "a" ) ) != NULL) /* open file */
    {
      fprintf( logfile, "%s:  error logging stopped\n\n", date );
      fclose( logfile );
      sprintf( string, "stop error logging in logfile %s", error_logfile );
      strcpy( error_logfile, "" );
      FctSetError( INFO, string, __FILE__, __LINE__ );
      ErrorMessage();
    }
    else
    {
      /* something is wrong with old log file */
      sprintf( string, "Cannot reopen logfile %s", error_logfile );
      strcpy( error_logfile, "" );
      FctSetError( ERROR, string, __FILE__, __LINE__ );
      ErrorMessage();
    }
  }

  error_logfile[FILENAMELENGTH - 1] = 0x0; /* make sure the name string is
    properly terminated, even if the input string is too long */
  strncpy( new_logfile, filename, FILENAMELENGTH - 1 );

  if ( new_logfile[0] )
  {
    sprintf( string, "start error logging in logfile %s", new_logfile );
    FctSetError( INFO, string, __FILE__, __LINE__ );
    ErrorMessage();
    strcpy( error_logfile, new_logfile );

    /* now start logging in new logfile */
    if (( logfile = fopen( error_logfile, "a+" ) ) != NULL) /* open file */
    {
       fprintf( logfile, "%s:  error logging started\n", date );
       fclose( logfile );
    }
    else
    {
      /* could not open logfile */
      sprintf( string, "Could not open logfile %s. Stop error logging", 
               error_logfile );
      strcpy( error_logfile, "" );
      FctSetError( ERROR, string, __FILE__, __LINE__ );
      ErrorMessage();
    }
  }
  else
  {
    strcpy( error_logfile, new_logfile );
  }

  return;
}


/* set error message */
void FctSetError( int el, const char *string, const char *file, int line )
{
  error_message[ERRORSTRINGLENGTH - 1] = 0x0; /* make sure the error string is 
    properly terminated, even if the input string is too long */
  filename[ERRORSTRINGLENGTH - 1] = 0x0; /* make sure the error string is 
    properly terminated, even if the input string is too long */
  strncpy( error_message, string, ERRORSTRINGLENGTH - 1 );
  strncpy( filename, file, ERRORSTRINGLENGTH - 1 );
  linenumber = line;

  errorlevel = el;

  return;
}


/* print error message */
void ErrorMessage()
{
  char string[ ERRORSTRINGLENGTH ];
  FILE *logfile;

  if ( FATAL - debuglevel <= errorlevel )
  {
    if ( long_format )
      fprintf( stderr, "\nIn line %d of file %s:\n", linenumber, filename );
    fprintf( stderr, "%s%s\n", errorcode[ errorlevel ], error_message );
  }

  if ( FATAL - logginglevel <= errorlevel )
  {
    if ( error_logfile[0] )
    {
      if (( logfile = fopen( error_logfile, "a" ) ) != NULL) /* open file */
      {
         if ( long_format )
           fprintf( logfile, "\nIn line %d of file %s:\n",
                             linenumber, filename );
         fprintf( logfile, "%s%s\n", errorcode[ errorlevel ], error_message );
         fclose( logfile );
      }
      else
      {
        /* could not open logfile */
        sprintf( string, "Could not open logfile %s. Stop error logging", 
                 error_logfile );
        ErrorLogfile( "" );
        FctSetError( ERROR, string, __FILE__, __LINE__ );
        ErrorMessage();
      }
    }
  }

  FctSetError( 0, "", "", 0 );

  return;
}
